"""Utilities used for testing
"""
import json
import struct
import wave
from pathlib import Path
from typing import Dict, List


def create_voice_files(base_path: Path, files: List[str], voice_name='voice', touch_only=False) -> Path:
    """Create voice files for testing

    Args:
        base_path (Path): Path to create voice folder at
        files (List[str]): Files to add to voice folder
        voice_name (str, optional): Voice name. Defaults to 'voice'.
        touch_only (bool, optional): Only touch voice files, don't use actual audio. Defaults to False.

    Returns:
        Path: _description_
    """
    audio_path = base_path.joinpath(voice_name)
    audio_path.mkdir()

    for file in files:
        filepath = audio_path.joinpath(file)
        if touch_only:
            filepath.touch()
            continue
        savefile = wave.open(str(filepath), "w")

        # channels, datasize (16 bit), sample rate, number of samples
        savefile.setparams((1, 2, 11025, 500, "NONE", "Uncompressed"))  # type: ignore
        savefile.writeframes(struct.pack('h', 1))
        savefile.close()

    return audio_path


def create_category_files(base_path: Path, category_files: Dict[str, List[str]], voice_name='voice') -> Path:
    """Create category files for testing

    Args:
        base_path (Path): Path to create category files in
        category_files (Dict[str, List[str]]): Files to create
        voice_name (str, optional): Voice name. Defaults to 'voice'.

    Returns:
        Path: Path to category files
    """
    audio_path = base_path.joinpath(voice_name)
    audio_path.mkdir()

    for category, files in category_files.items():
        create_voice_files(
            base_path=audio_path,
            files=files,
            voice_name=category
        )

    return audio_path


def create_info(info: dict, audio_path: Path):
    """Create info file for testing

    Args:
        info (dict): Info contents
        audio_path (Path): Path to create info file in
    """
    info_dir = Path(audio_path).joinpath("info/")

    info_dir.mkdir(parents=True, exist_ok=True)

    info_file = info_dir.joinpath("info.json")

    with open(info_file, 'w', encoding='UTF-8') as file:
        json.dump(info, file)



"""Manages multiple voices
"""
import argparse
import logging
import sys
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, List, Optional, Union

import sqlalchemy

from hlvox.voice import MultiVoice, SingleVoice, Voice

log = logging.getLogger(__name__)


class DuplicateVoice(Exception):
    """Raised when duplicate voices are found
    """


@dataclass
class RemoteDatabaseInfo:
    """Database connection information if using remote database (sql)
    """
    name: str
    url: str
    port: int
    username: str
    password: str


@dataclass
class LocalDatabaseInfo:
    """Database info if using local file-based database (sqlite)
    """
    base_path: Path


class Manager:
    """Manages multiple voices
    """

    def __init__(
        self,
        voices_path: Union[Path, str],
        database_info: Optional[Union[LocalDatabaseInfo, RemoteDatabaseInfo]]
    ):
        self.voices_path = Path(voices_path)
        self._database_info = database_info

        # This strangeness is to make mypy happy. There is probably a cleaner way to do it.
        single_voices = self._load_voices(self.voices_path)
        multi_voice = self._create_multi_voice(single_voices)
        voices: Dict[str, Union[MultiVoice, SingleVoice]] = {}
        voices.update(single_voices)
        voices['multi'] = multi_voice
        self.voices = voices

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.exit()

    def _create_db_path(self, databases_path: Path, voice_name: str) -> Path:
        db_path = databases_path / voice_name
        db_path.mkdir(parents=True, exist_ok=True)
        return db_path

    def _create_db(self, database_info: Union[LocalDatabaseInfo, RemoteDatabaseInfo], voice_name: str) -> sqlalchemy.engine.Engine:
        dbi = database_info
        if isinstance(dbi, LocalDatabaseInfo):
            path = self._create_db_path(databases_path=dbi.base_path, voice_name=voice_name)
            return sqlalchemy.create_engine(f'sqlite:///{path}/db.sqlite')
        return sqlalchemy.create_engine(f'postgresql+psycopg2://{dbi.username}:{dbi.password}@{dbi.url}:{dbi.port}/{dbi.name}', pool_pre_ping=True)

    def _load_voices(self, path: Path) -> Dict[str, SingleVoice]:
        voices = {}
        voice_folders = list(x for x in path.iterdir() if x.is_dir())
        for voice_folder in voice_folders:
            voice_name = voice_folder.name.lower()
            if voice_name in voices:
                raise DuplicateVoice('Duplicate voice name found')

            database = None
            if self._database_info is not None:
                database = self._create_db(database_info=self._database_info, voice_name=voice_name)
            new_voice = SingleVoice(
                name=voice_name, path=voice_folder, database=database)
            voices[new_voice.name] = new_voice
        return voices

    def _create_multi_voice(self, voices: Dict[str, SingleVoice]) -> MultiVoice:
        database = None
        if self._database_info is not None:
            database = self._create_db(database_info=self._database_info, voice_name='multi')
        return MultiVoice(
            voices=voices,
            database=database,
        )

    def get_voice_names(self) -> List[str]:
        """Gets names of available voices

        Returns:
            list -- list of voice name strings
        """

        voice_names = list(self.voices.keys())
        voice_names.sort()
        return voice_names

    def get_voice(self, name: str) -> Optional[Voice]:
        """Get voice of requested name

        Args:
            name ({string}): name of voice to get

        Returns:
            {voxvoice}: requested voice
        """
        if name in self.voices:
            return self.voices[name]
        return None

    def exit(self):
        """Exit all loaded voices
        """
        for voice in self.voices.values():
            voice.exit()


if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)
    parser = argparse.ArgumentParser(
        description='Generate a sentence using voices')
    parser.add_argument('-s', '--voices-dir', type=str, required=True,
                        help='Path to folder with voice audio file folders')
    parser.add_argument('-f', '--format', type=str, required=False,
                        default='wav', help='Audio format to export as')
    parser.add_argument('voice', type=str)
    parser.add_argument('sentence', type=str)
    args = parser.parse_args()

    voices_dir = Path(args.voices_dir)
    if not voices_dir.is_dir():
        log.error('Voices dir at %s does not exist!', voices_dir)
        sys.exit(1)

    manager = Manager(
        voices_path=voices_dir,
        database_info=None,
    )

    loaded_voice = manager.get_voice(args.voice)
    if loaded_voice is None:
        log.error("Voice %s was not found", loaded_voice)
        sys.exit(1)

    sentence = loaded_voice.generate_audio(args.sentence)
    if sentence is None or sentence.audio is None:
        log.error('Cannot generate %s: %s', sentence.sentence, sentence)
        sys.exit(1)

    # Paths can't have : in them, so replace with % as a stand-in
    sanitized_sentence = sentence.sentence.replace(':', '%')
    output_path = Path.cwd().joinpath(f"{sanitized_sentence}.{args.format}")

    log.info('Exporting to %s', output_path)
    sentence.audio.export(output_path, format=args.format)



"""Test voice manager
"""
# pylint: disable=missing-function-docstring,disable=missing-class-docstring
# pylint: disable=too-few-public-methods
import sys
from pathlib import Path

import pytest

import hlvox

from . import utils as th

test_files = {
    "v1": ["hello.wav"],
    "v2": ["hello.wav"],
    "v3": ["hello.wav"]
}


@pytest.fixture(name="manager")
def manager_fixture(tmp_path: Path):
    voices_dir = tmp_path.joinpath("voices")
    voices_dir.mkdir()
    for voice, words in test_files.items():
        th.create_voice_files(voices_dir, words, voice)
    manager = hlvox.Manager(voices_path=voices_dir, database_info=None)
    yield manager
    manager.exit()


class TestVoiceImport():
    def test_voice_list(self, manager: hlvox.Manager):
        expected_voices = list(test_files.keys())
        expected_voices.append("multi")
        expected_voices.sort()
        assert manager.get_voice_names() == expected_voices


class TestGetVoice():
    def test_get_voice(self, manager: hlvox.Manager):
        voice = manager.get_voice("v1")

        assert voice is not None
        assert voice.name == "v1"

    def test_get_wrong_voice(self, manager: hlvox.Manager):
        voice = manager.get_voice("nope")

        assert not voice


class TestDuplicatevoice():
    @pytest.mark.skipif(sys.platform == "win32", reason="Windows is not case sensitive so duplicates cant exist")
    def test_case_diff(self, tmp_path: Path):
        voices_dir = tmp_path.joinpath("voices")
        voices_dir.mkdir()
        th.create_voice_files(voices_dir, ["hello.wav"], "AHH")
        th.create_voice_files(voices_dir, ["hello.wav"], "ahh")
        th.create_voice_files(voices_dir, ["hello.wav"], "aHH")

        with pytest.raises(hlvox.DuplicateVoice):
            hlvox.Manager(voices_path=voices_dir,
                          database_info=None)



"""Test voice
"""
# NOTE: since these tests use sqlite as their DB backend, indexes for results
# start at 1. This may differ if using postgresql in actual use.

# pylint: disable=missing-function-docstring,disable=missing-class-docstring
# pylint: disable=too-few-public-methods
# TODO: This isn't great and implies some re-architecting is needed
# pylint: disable=protected-access
import sys
from pathlib import Path

import pytest
import sqlalchemy

import hlvox

from . import utils as th

# Stand-in files for testing
normal_files = ["hello.wav", "my.wav", "name.wav", "is.wav", "vox.wav"]
inconst_format_files = ["hello.mp3", "my.wav", "name", "is.wav", "vox.mp4"]
no_format_files = ["imatextfile", "whatami"]
alph_files = ["a.wav", "b.wav", "c.wav"]
category_files = {'bad': ["stop.wav", "bad.wav",
                          "no.wav"], 'good': ["go.wav", "yes.wav"]}


class TestFileHandling():
    def test_empty_files(self, tmp_path: Path):
        voice_dir = th.create_voice_files(tmp_path, [])
        with pytest.raises(hlvox.NoWordsFound):
            hlvox.SingleVoice(name='test', path=voice_dir,
                              database=None)

    def test_inconsistent_format(self, tmp_path: Path):
        voice_dir = th.create_voice_files(tmp_path, inconst_format_files)
        with pytest.raises(hlvox.InconsistentAudioFormats):
            hlvox.SingleVoice(name='test', path=voice_dir,
                              database=None)

    def test_no_format(self, tmp_path: Path):
        voice_dir = th.create_voice_files(tmp_path, no_format_files)
        with pytest.raises(hlvox.NoAudioFormatFound):
            hlvox.SingleVoice(name='test', path=voice_dir,
                              database=None)

    def test_audio_format(self, tmp_path: Path):
        voice_dir = th.create_voice_files(tmp_path, normal_files)
        with hlvox.SingleVoice(name='test', path=voice_dir, database=None) as unit:
            assert unit.get_audio_format() == "wav"


class TestDictContents():
    def test_basic_dict(self, tmp_path: Path):
        voice_dir = th.create_voice_files(tmp_path, normal_files)

        expected_names = [name[:-4] for name in normal_files]
        expected_names.sort()

        with hlvox.SingleVoice(name='test', path=voice_dir, database=None) as unit:
            assert expected_names == unit.words

    def test_alphab(self, tmp_path: Path):
        voice_dir = th.create_voice_files(tmp_path, alph_files)

        expected_names = ["a", "b", "c"]

        with hlvox.SingleVoice(name='test', path=voice_dir, database=None) as unit:
            assert expected_names == unit.words

    def test_caps(self, tmp_path: Path):
        words = [
            "Cap.wav",
            "Cappa.wav",
        ]
        voice_dir = th.create_voice_files(tmp_path, words)

        expected_names = ["cap", "cappa"]

        with hlvox.SingleVoice(name='test', path=voice_dir, database=None) as unit:
            assert expected_names == unit.words

    def test_categories(self, tmp_path: Path):
        voice_dir = th.create_category_files(tmp_path, category_files)
        expected_names = [item[:-4]
                          for sublist in category_files.values() for item in sublist]
        expected_names.sort()
        with hlvox.SingleVoice(name='test', path=voice_dir, database=None) as unit:
            assert expected_names == unit.words

            for category, words in category_files.items():
                category_words = [word[:-4] for word in words]
                category_words.sort()
                assert unit.categories[category] == category_words

    @pytest.mark.skipif(sys.platform == "win32", reason="Windows is not case sensitive so duplicates cant exist")
    def test_duplicates(self, tmp_path: Path):
        words = [
            "Cap.wav",
            "cAP.wav",
        ]
        voice_dir = th.create_voice_files(tmp_path, words, touch_only=True)

        with pytest.raises(hlvox.DuplicateWords):
            hlvox.SingleVoice(name='test', path=voice_dir,
                              database=None)


class TestDatabase():
    def test_reuse(self, tmp_path: Path):
        voice_dir = th.create_voice_files(tmp_path, normal_files)
        database = sqlalchemy.create_engine(f'sqlite:///{tmp_path}/db.sqlite')
        voice = hlvox.SingleVoice(
            name='test', path=voice_dir, database=database)

        voice.generate_audio(sentence="hello hello")
        voice.exit()

        voice = hlvox.SingleVoice(
            name='test', path=voice_dir, database=database)
        assert voice.get_generated_sentences() == ["hello hello"]


@pytest.fixture(name="voice")
def voice_fixture(tmp_path: Path):
    voice_dir = th.create_voice_files(tmp_path, normal_files)
    database = sqlalchemy.create_engine(f'sqlite:///{tmp_path}/db.sqlite')
    voice = hlvox.SingleVoice(
        name='test', path=voice_dir, database=database)
    yield voice
    voice.exit()


class TestSayableUnsayable():
    def test_empty_sent(self, voice: hlvox.SingleVoice):
        ret_say, ret_unsay = voice.get_sayable_unsayable([])

        assert ret_say == []
        assert ret_unsay == []

    def test_simple_sent(self, voice: hlvox.SingleVoice):
        words = ["hello"]
        ret_say, ret_unsay = voice.get_sayable_unsayable(words)

        assert ret_say == ["hello"]
        assert ret_unsay == []

    def test_duplicates(self, voice: hlvox.SingleVoice):
        words = ["hello", "hello", "world", "world", ",",
                 ",", ".", ".", "duplicates!", "duplicates"]

        ret_say, ret_unsay = voice.get_sayable_unsayable(words)

        assert not set(ret_say) ^ set(["hello", ",", "."])
        assert not set(ret_unsay) ^ set(["world", "duplicates", "duplicates!"])

    def test_comp_sent(self, voice: hlvox.SingleVoice):
        words = ["hello", ",", "world", ".", "vox",
                 "can't", "say", "some", "of", "this", "."]

        ret_say, ret_unsay = voice.get_sayable_unsayable(words)

        assert not set(ret_say) ^ set(["hello", ",", "vox", "."])
        assert not set(ret_unsay) ^ set(
            ["world", "can't", "say", "some", "of", "this"]
        )

    def test_dup_punct(self, voice: hlvox.SingleVoice):
        words = ["hello", ".", ".", ".", "world"]

        ret_say, ret_unsay = voice.get_sayable_unsayable(words)

        assert not set(ret_say) ^ set(["hello", "."])
        assert not set(ret_unsay) ^ set(["world"])


class TestSentenceGeneration():
    def test_empty_sent(self, voice: hlvox.SingleVoice):
        ret = voice.generate_audio("")

        assert ret.sentence == ""
        assert ret.sayable == []
        assert ret.unsayable == []
        assert ret.audio is None

        assert voice.get_generated_sentences_dict() == {}

    def test_unsayable_sent(self, voice: hlvox.SingleVoice):
        ret = voice.generate_audio("whatthefuckdidyoujustsaytome")

        assert ret.sentence == ""
        assert ret.sayable == []
        assert ret.unsayable == ["whatthefuckdidyoujustsaytome"]
        assert ret.audio is None

        assert voice.get_generated_sentences_dict() == {}

    def test_sayable_sent(self, voice: hlvox.SingleVoice):
        sentence = "hello, my name is vox"
        ret = voice.generate_audio(sentence)

        assert ret.sentence == "hello , my name is vox"
        assert ret.sayable == [",", "hello", "is", "my", "name", "vox"]
        assert ret.unsayable == []
        assert ret.audio is not None

        assert voice.get_generated_sentences_dict() == {
            1: "hello , my name is vox"}

    def test_duplicate_sent(self, voice: hlvox.SingleVoice):
        voice.generate_audio("hello hello")
        voice.generate_audio("hello hello")

        assert voice.get_generated_sentences_dict() == {1: "hello hello"}
        assert len(voice._get_generated_sentences_list()) == 1

    def test_duplicate_words(self, voice: hlvox.SingleVoice):
        ret = voice.generate_audio("hello hello hello")

        assert ret.sentence == "hello hello hello"
        assert ret.sayable == ["hello"]
        assert ret.unsayable == []
        assert ret.audio is not None

    def test_dup_punct(self, voice: hlvox.SingleVoice):
        ret = voice.generate_audio("hello... hello")

        assert ret.sentence == "hello . . . hello"
        assert ret.sayable == [".", "hello"]
        assert ret.unsayable == []
        assert ret.audio is not None

    def test_multiple_sent(self, voice: hlvox.SingleVoice):
        voice.generate_audio("hello hello")
        voice.generate_audio("vox hello")

        assert voice.get_generated_sentences_dict() == {
            1: "hello hello",
            2: "vox hello",
        }

    def test_dry_run(self, voice: hlvox.SingleVoice):
        ret = voice.generate_audio("hello", dry_run=True)
        assert ret.audio is None

    def test_one_word(self, voice: hlvox.SingleVoice):
        # We don't bother saving single words to the database
        voice.generate_audio("hello")

        assert voice.get_generated_sentences_dict() == {}


class TestGetGeneratedSentences():
    def test_no_sentences(self, voice: hlvox.SingleVoice):
        ret = voice.get_generated_sentences()

        assert ret == []

    def test_single_sentences(self, voice: hlvox.SingleVoice):
        voice.generate_audio("hello hello")

        ret = voice.get_generated_sentences()

        assert ret == ["hello hello"]

    def test_multiple_sentences(self, voice: hlvox.SingleVoice):
        voice.generate_audio("hello hello")
        voice.generate_audio("vox hello")

        ret = voice.get_generated_sentences()

        assert ret == ["hello hello", "vox hello"]


class TestGetGeneratedSentencesDict():
    def test_no_sentences(self, voice: hlvox.SingleVoice):
        ret = voice.get_generated_sentences_dict()

        assert ret == {}

    def test_single_sentences(self, voice: hlvox.SingleVoice):
        voice.generate_audio("hello hello")

        ret = voice.get_generated_sentences_dict()

        assert ret == {1: "hello hello"}

    def test_multiple_sentences(self, voice: hlvox.SingleVoice):
        voice.generate_audio("hello hello")
        voice.generate_audio("vox hello")

        ret = voice.get_generated_sentences_dict()

        assert ret == {1: "hello hello", 2: "vox hello"}


@pytest.fixture(name="multi_voice")
def multi_voice_fixture(tmp_path: Path):
    norm_voice_dir = th.create_voice_files(
        tmp_path, normal_files, voice_name='norm')
    alph_voice_dir = th.create_voice_files(
        tmp_path, alph_files, voice_name='alph')
    alph_2_voice_dir = th.create_voice_files(
        tmp_path, alph_files, voice_name='alph2')
    norm_voice = hlvox.SingleVoice(
        name='norm',
        path=norm_voice_dir,
        database=sqlalchemy.create_engine(f'sqlite:///{tmp_path}/norm.sqlite')
    )
    alph_voice = hlvox.SingleVoice(
        name='alph',
        path=alph_voice_dir,
        database=sqlalchemy.create_engine(f'sqlite:///{tmp_path}/alph.sqlite')
    )
    alph_2_voice = hlvox.SingleVoice(
        name='alph2',
        path=alph_2_voice_dir,
        database=sqlalchemy.create_engine(f'sqlite:///{tmp_path}/alph2.sqlite')
    )
    voices = {
        'norm': norm_voice,
        'alph': alph_voice,
        'alph2': alph_2_voice,
    }
    multi_voice = hlvox.MultiVoice(
        voices=voices,
        database=sqlalchemy.create_engine(f'sqlite:///{tmp_path}/multi.sqlite'),
    )
    yield multi_voice
    multi_voice.exit()
    for voice in voices.values():
        voice.exit()


EXPECTED_MULTI_VOICE_WORDS = [
    'norm:hello', 'norm:is', 'norm:my', 'norm:name',  'norm:vox',
    'alph:a', 'alph:b', 'alph:c',
    'alph2:a', 'alph2:b', 'alph2:c',
]

# Multi-voice tests


class TestMultiWords:
    def test_normal(self, multi_voice: hlvox.MultiVoice):
        assert multi_voice.words == EXPECTED_MULTI_VOICE_WORDS


class TestMultiProcessSentence:
    # Just want to check that punctuation works as expected
    def test_normal(self, multi_voice: hlvox.MultiVoice):
        words = ["norm:hello", ",", "norm:my",
                 "norm:name", "alph:a,", "alph2:b", "."]
        exp_words = ["norm:hello", ",", "norm:my",
                     "norm:name", "alph:a", ",", "alph2:b", "."]
        ret = multi_voice.process_sentence(words)
        assert ret == exp_words


class TestMultiSayableUnsayable:
    def test_empty(self, multi_voice: hlvox.MultiVoice):
        sayable, unsayable = multi_voice.get_sayable_unsayable(words=[])
        assert not sayable
        assert not unsayable

    def test_all_sayable(self, multi_voice: hlvox.MultiVoice):
        words = ["norm:hello", "norm:my", "alph:a", "alph2:a"]
        sayable, unsayable = multi_voice.get_sayable_unsayable(words=words)
        words.sort()
        assert sayable == words
        assert not unsayable

    def test_some_unsayable(self, multi_voice: hlvox.MultiVoice):
        exp_unsayable = ["norm:nope", "alph:butter", "alph2:404"]
        exp_sayable = ["norm:vox", "alph:c", "alph2:c"]
        words = exp_sayable + exp_unsayable
        sayable, unsayable = multi_voice.get_sayable_unsayable(words=words)
        exp_sayable.sort()
        exp_unsayable.sort()
        assert sayable == exp_sayable
        assert unsayable == exp_unsayable

    def test_punctuation(self, multi_voice: hlvox.MultiVoice):
        words = ["norm:hello", "norm:,", "alph:a", "alph2:."]
        words.sort()
        sayable, unsayable = multi_voice.get_sayable_unsayable(words=words)
        assert sayable == words
        assert not unsayable


class TestMultiWordVoiceAssignment:
    def test_empty(self, multi_voice: hlvox.MultiVoice):
        assert not multi_voice._get_word_voice_assignment(words=[])

    def test_no_initial_voice(self, multi_voice: hlvox.MultiVoice):
        words = ["hello", "norm:my"]

        with pytest.raises(hlvox.NoVoiceSpecified):
            multi_voice._get_word_voice_assignment(words=words)

    def test_normal(self, multi_voice: hlvox.MultiVoice):
        voices = multi_voice._voices
        words = ["norm:hello", "my", "alph:a", "a",
                 "alph2:a", "norm:vox", "is", "alph:b"]
        exp_words_and_voices = [
            ("hello", voices["norm"]),
            ("my", voices["norm"]),
            ("a", voices["alph"]),
            ("a", voices["alph"]),
            ("a", voices["alph2"]),
            ("vox", voices["norm"]),
            ("is", voices["norm"]),
            ("b", voices["alph"]),
        ]

        assert multi_voice._get_word_voice_assignment(
            words=words) == exp_words_and_voices

    def test_punctuation(self, multi_voice: hlvox.MultiVoice):
        voices = multi_voice._voices
        words = ["norm:hello", ",", "alph:a", "a",
                 ".", "alph2:a", "norm:vox", "is", "alph:b"]
        exp_words_and_voices = [
            ("hello", voices["norm"]),
            (",", voices["norm"]),
            ("a", voices["alph"]),
            ("a", voices["alph"]),
            (".", voices["alph"]),
            ("a", voices["alph2"]),
            ("vox", voices["norm"]),
            ("is", voices["norm"]),
            ("b", voices["alph"]),
        ]

        assert multi_voice._get_word_voice_assignment(
            words=words) == exp_words_and_voices


class TestMultiCombinedVoiceSentences:
    def test_empty(self, multi_voice: hlvox.MultiVoice):
        assert not multi_voice.get_combined_voice_sentences(
            words_and_voices=[])

    def test_single(self, multi_voice: hlvox.MultiVoice):
        voices = multi_voice._voices
        words_and_voices = [
            ("hello", voices["norm"])
        ]

        exp = [(voices["norm"], ["hello"])]
        assert multi_voice.get_combined_voice_sentences(
            words_and_voices=words_and_voices) == exp

    def test_multiple(self, multi_voice: hlvox.MultiVoice):
        voices = multi_voice._voices
        words_and_voices = [
            ("hello", voices["norm"]),
            ("my", voices["norm"]),
            ("a", voices["alph"]),
            ("a", voices["alph"]),
            ("a", voices["alph2"]),
            ("vox", voices["norm"]),
            ("is", voices["norm"]),
            ("b", voices["alph"]),
        ]

        exp = [
            (voices["norm"], ["hello", "my"]),
            (voices["alph"], ["a", "a"]),
            (voices["alph2"], ["a"]),
            (voices["norm"], ["vox", "is"]),
            (voices["alph"], ["b"]),
        ]
        assert multi_voice.get_combined_voice_sentences(
            words_and_voices=words_and_voices) == exp


class TestMultiGenerateAudio:
    def test_normal(self, multi_voice: hlvox.MultiVoice):
        sentence = "norm:hello, my name is vox. alph:a b c"
        ret = multi_voice.generate_audio(sentence=sentence)

        exp_sentence = "norm:hello , my name is vox . alph:a b c"
        assert ret.sentence == exp_sentence
        # Sentence should not be saved into the individual voice's database
        assert len(multi_voice._voices["norm"].get_generated_sentences()) == 0
        assert len(multi_voice._voices["alph"].get_generated_sentences()) == 0


class TestDifficultSentences:
    def test_ellipse(self, multi_voice: hlvox.MultiVoice):
        sentence = "norm:hello norm:. norm:. norm:."
        ret = multi_voice.generate_audio(sentence=sentence)

        exp_sentence = "norm:hello . . ."
        assert ret.sentence == exp_sentence

        sentence = "norm:hello..."
        ret = multi_voice.generate_audio(sentence=sentence)
        exp_sentence = "norm:hello . . ."
        assert ret.sentence == exp_sentence

    def test_start_ellipse(self, multi_voice: hlvox.MultiVoice):
        sentence = "norm:."
        ret = multi_voice.generate_audio(sentence=sentence)

        exp_sentence = "norm:."
        assert ret.sentence == exp_sentence

        sentence = "norm:. . ."
        ret = multi_voice.generate_audio(sentence=sentence)

        exp_sentence = "norm:. . ."
        assert ret.sentence == exp_sentence

        sentence = "norm:..."
        ret = multi_voice.generate_audio(sentence=sentence)
        assert ret.sentence == exp_sentence



"""Defines a unique voice
"""
import argparse
import json
import logging
import sys
from collections import defaultdict
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, List, Optional, Tuple

import sqlalchemy
from pydub import AudioSegment

log = logging.getLogger(__name__)


# How much delay should be added in place of punctuation (in milliseconds)
PUNCTUATION_TIMING = {
    ',': 250,
    '.': 500,
}

DB_NAME = 'db.json'


class NoWordsFound(Exception):
    """Raised when a voice has no words
    """


class DuplicateWords(Exception):
    """Raised when a voice has duplicate words
    """


class InconsistentAudioFormats(Exception):
    """Raised when words have inconsistent audio formats
    """


class NoAudioFormatFound(Exception):
    """Raised when no audio format can be found
    """


class FailedToSplit(Exception):
    """Raised when a sentence cannot be split
    """


class NoVoiceSpecified(Exception):
    """Raised when no voice is specified
    """


class NoDatabaseSpecified(Exception):
    """No database connection was specified during init
    """


@dataclass
class Sentence:
    """Represents a sentence and it's parts
    """
    sentence: str
    sayable: List[str]
    unsayable: List[str]
    audio: Optional[AudioSegment] = None


@dataclass
class DatabaseConnection:
    """Stores info related to database connection
    """
    engine: sqlalchemy.engine.Engine
    metadata: sqlalchemy.MetaData
    sentence_table: sqlalchemy.Table


class Voice:
    """Base class for Voice-like interfaces.
    Intended to involve generation of audio
    files from some source (files, web, etc).
    """

    def __init__(
        self,
        name: str,
        database: Optional[sqlalchemy.engine.Engine],
    ):
        self.name = name

        self._db: Optional[DatabaseConnection] = None
        if database is not None:
            metadata = sqlalchemy.MetaData()
            sentence_table = sqlalchemy.Table(
                self.name,
                metadata,
                sqlalchemy.Column("id", sqlalchemy.Integer, primary_key=True),
                sqlalchemy.Column("sentence", sqlalchemy.String, unique=True, nullable=False),
            )
            metadata.create_all(database)
            self._db = DatabaseConnection(
                metadata=metadata,
                sentence_table=sentence_table,
                engine=database,
            )
        self.words: List[str] = []
        self.categories: Dict[str, List[str]] = {}

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.exit()

    def exit(self):
        """Clean up and close voice
        """
        if self._db is not None:
            self._db.engine.dispose()

    def _insert_sentence_into_db(self, sentence: str):
        if self._db is None:
            raise NoDatabaseSpecified
        ins = self._db.sentence_table.insert().values(sentence=sentence)
        with self._db.engine.connect() as conn:
            conn.execute(ins)

    def _sentence_exists(self, sentence: str) -> bool:
        if self._db is None:
            raise NoDatabaseSpecified

        sel = self._db.sentence_table.select().where(self._db.sentence_table.c.sentence == sentence)
        with self._db.engine.connect() as conn:
            result = conn.execute(sel)
            # TODO: there should be a way to use `.Count()` here
            return bool(result.all())

    def _get_generated_sentences_list(self) -> list[str]:
        if self._db is None:
            raise NoDatabaseSpecified
        sel = self._db.sentence_table.select()
        with self._db.engine.connect() as conn:
            result = conn.execute(sel)
            return [r.sentence for r in result]

    def _get_generated_sentences_dict(self) -> dict[int, str]:
        if self._db is None:
            raise NoDatabaseSpecified
        sel = self._db.sentence_table.select()
        with self._db.engine.connect() as conn:
            result = conn.execute(sel)
            return {r.id: r.sentence for r in result}

    def _get_sentence_info(self, words: List[str]) -> Tuple[Sentence, List[str]]:
        """Get basic sentence info for a given
        split sentence.

        Args:
            words (List[str]): Words of sentence split into array

        Returns:
            Tuple[Sentence, List[str]]: Sentence info and sayable sentence array
        """
        sayable_words, unsayable_worlds = self.get_sayable_unsayable(words)
        sayable_sent_arr = self._get_sayable_sentence_arr(
            words, sayable_words)
        sayable_sent_str = self._create_sentence_string(sayable_sent_arr)

        sentence_info = Sentence(
            sentence=sayable_sent_str,
            sayable=sayable_words,
            unsayable=unsayable_worlds,
            audio=None,
        )
        return sentence_info, sayable_sent_arr

    def generate_audio_from_array(self, words: List[str], dry_run=False, save_to_db=True) -> Sentence:
        """Generates audio segment from sentence array.

        Args:
            words (List[str]): Words to try and turn into audio segment.
            dry_run (bool, optional): Skip actual segment generation. Defaults to False.

        Returns:
            Sentence: Sentence with audio segment.
        """
        sentence_info, sayable_sentence_array = self._get_sentence_info(
            words=words)

        if dry_run:
            return sentence_info

        log.debug("Generating %s", sentence_info.sentence)

        # Only create sentence if there are words to put in it
        if len(sentence_info.sayable) == 0:
            log.warning(
                "Can't say any words in %s, not generating", sentence_info.sentence)
            return sentence_info

        # Only bother inserting a sentence into the database if there is more than one word in it
        # TODO: test save_to_db
        if save_to_db and len(words) > 1:
            if not self._sentence_exists(sentence=sentence_info.sentence):
                self._insert_sentence_into_db(sentence=sentence_info.sentence)

        words_audio = self._create_audio_segments(sayable_sentence_array)

        sentence_info.audio = self.assemble_audio_segments(words_audio)

        return sentence_info

    def _create_audio_segments(
        self,
        word_array: List[str],  # pylint: disable=unused-argument
    ) -> List[AudioSegment]:
        """Create audio segments for each entry in an array of words.

        Args:
            word_array (List[str]): Words to turn into audio segments.

        Returns:
            List[AudioSegment]: Audio segments.
        """
        return []

    def generate_audio(self, sentence: str, dry_run=False) -> Sentence:
        """Generates audio from the given sentence

        Args:
            sentence (string): Sentence string to be generated
            dry_run (bool, optional): Don't generate audio. Defaults to False.
        Returns:
            Sentence: Information about generated sentence.
        """
        log.info("Asked to generate %s", sentence)
        split_sentence = self._split_sentence(sentence)
        proc_sentence = self.process_sentence(split_sentence)
        return self.generate_audio_from_array(
            words=proc_sentence,
            dry_run=dry_run,
        )

    @staticmethod
    def _split_sentence(sentence: str) -> List[str]:
        return sentence.lower().rstrip().split(" ")

    @staticmethod
    def process_sentence(split_sent: List[str]) -> List[str]:
        """
        Takes a normally formatted sentence and breaks it into base elements

        Args:
            split_sent (List[str]): words in sentence

        Returns:
            List[str]: array of elements in sentence
        """
        # TODO: This could use some rethinking. Should be easier to just always break punctuation marks
        # into their own elements, rather than selectively only dealing with trailing ones.
        log.info("Processing sentence '%s'", split_sent)

        # Pull out punctuation
        reduced_sent = []
        for item in split_sent:
            # find first punctuation mark, if any
            first_punct: Optional[str] = None
            try:
                first_punct = next(
                    (punct for punct in PUNCTUATION_TIMING if punct in item))
            except StopIteration:
                pass

            if first_punct:
                # Get its index
                first_punct_ind = item.find(first_punct)

                # Special case: If this is a multi voice sentence,
                # we don't want to rip the voice definition out of a singe-punctuation
                # mark. IE vox:.
                # TODO: This is a bit hacky. Would be great if this method doesn't
                # have to know about multi-voice syntax.
                if first_punct_ind >= 2 and item[first_punct_ind - 1] == ':':
                    reduced_sent.append(item[:first_punct_ind + 1])
                    if len(item) >= first_punct_ind:
                        first_punct_ind += 1
                else:
                    # Add everything before punct (the word, if any)
                    if item[:first_punct_ind]:
                        reduced_sent.append(item[:first_punct_ind])

                # Add all the punctuation if its actually punctuation
                # TODO: Figure out if I want to deal with types like ".hello" throwing out all the characters after the period.
                for punct in item[first_punct_ind:]:
                    if punct in PUNCTUATION_TIMING:
                        reduced_sent.append(punct)

            else:
                reduced_sent.append(item)

        # Clean blanks from reduced_sent
        if '' in reduced_sent:
            reduced_sent = [value for value in reduced_sent if value != '']

        log.info("Sentence processed: '%s'", reduced_sent)
        return reduced_sent

    def get_sayable_unsayable(self, words: List[str]) -> Tuple[List[str], List[str]]:
        """Get words that are sayable or unsayable
        from a list of words.

        Args:
            words (List[str]): Words to check.

        Returns:
            Tuple[List[str], List[str]]: Sayable and unsayable words.
        """
        # TODO: This shouldn't need two separate processings of the same sentence. Sets, people. Sets!
        sayable_words = self.words
        sayable_words_set = set(sayable_words)
        sayable_words_set.update(list(PUNCTUATION_TIMING.keys()))

        words_set = set((dict.fromkeys(words)))  # removes duplicates

        unsayable_set = words_set - sayable_words_set
        sayable_set = words_set - unsayable_set
        unsayable = list(unsayable_set)
        unsayable.sort()
        sayable = list(sayable_set)
        sayable.sort()
        return sayable, unsayable

    def _get_sayable_sentence_arr(self, words: List[str], sayable_words: List[str]) -> List[str]:
        """Removes words from sentence array that are not sayable.

        Args:
            words (List[str]): Array of words in sentence, in order.
            sayable_words (List[str]): Words from sentence that can actually be said.

        Returns:
            List[str]: Words in sentence that are sayable, in order.
        """
        # TODO: This is just a simple set operation. Function probably isn't needed. At least change to using a set.
        return [word for word in words if word in sayable_words]

    def _create_sentence_string(self, words: List[str]) -> str:
        """Joins sentence array into a string.

        Args:
            words (List[str]): Words in sentence, in order.

        Returns:
            str: Sentence string.
        """
        if len(words) == 1:
            return words[0]
        return " ".join(words)

    def get_generated_sentences(self) -> List[str]:
        """Gets the previously generated sentence strings

        Returns:
            List[str]: List of sentence strings generated previously
        """
        return self._get_generated_sentences_list()

    def get_generated_sentences_dict(self) -> Dict[int, str]:
        """Gets the previously generated sentence strings
        along with their corresponding ID in the database

        Returns:
            Dict[int, str]: Dict of sentence and id pairs
        """
        return self._get_generated_sentences_dict()

    @staticmethod
    def assemble_audio_segments(segments: List[AudioSegment]) -> AudioSegment:
        """Assemble audio segments into one audio segment.

        Args:
            segments (List[AudioSegment]): Segments to assemble.

        Returns:
            AudioSegment: Assembled audio segment.
        """
        # We set all audio segments to the lowest common frame rate
        # to avoid some really ugly artifacting when a low frame rate
        # clip is appended to a high frame rate one.
        frame_rates = [word.frame_rate for word in segments]
        frame_rate = min(frame_rates)

        sentence_audio = segments.pop(0)
        sentence_audio = sentence_audio.set_frame_rate(frame_rate)
        for word_audio in segments:
            word_audio = word_audio.set_frame_rate(frame_rate)
            sentence_audio = sentence_audio + word_audio

        return sentence_audio


class SingleVoice(Voice):
    """Comprises all information and methods
    needed to index a folder of voice audio files
    and generate audio from them given a sentence string.
    """

    def __init__(self, name: str, path: Path, database: Optional[DatabaseConnection]):
        """
        Args:
            name: Name of voice
            path (Path): Path to folder of voice audio files.
            database (Optional[DatabaseConnection]): Database connection information.
                If none provided, no database will be used and no data will persist.
        """
        super().__init__(name=name, database=database)
        self.path = path

        self.info_path = self.path.joinpath("info/")
        self.info_name = "info.json"

        self._word_dict, self.categories = self._build_word_dict(self.path)
        self._audio_format = self._find_audio_format(
            self._word_dict)  # TODO: Use properies?

        self.words = self._get_words()

        self._read_info(self.info_path, self.info_name)

    def _build_word_dict(self, path: Path) -> Tuple[Dict[str, Path], Dict[str, List[str]]]:
        """Builds dictionary of all available words and categories.

        Args:
            path (Path): Path to folder of voice audio files, or folders of voices files.

        Raises:
            DuplicateWords: Raised if there are duplicate filenames present.
            NoWordsFound: Raised if no words are found.

        Returns:
            Tuple[Dict[str, Path], Dict[str, List[str]]]: Dict of {filepath: word} associations and {category: [words]}.
        """
        word_dict = {}
        categories = defaultdict(list)

        for word_path in path.glob("**/*"):
            if word_path.is_dir():
                continue
            if word_path.parent.name == 'info':
                continue
            word = word_path
            name = str(word.stem).lower()
            if name in word_dict:
                raise DuplicateWords(f"Word {name} is duplicated")
            category = ''
            if word.parent != path:
                category = word.parent.name

            word_dict[name] = word
            if category:
                categories[category].append(name)
                # This is probably bad
                categories[category].sort()

        if len(word_dict) == 0:
            log.error("No words found")
            raise NoWordsFound

        return word_dict, categories

    def _read_info(self, path: Path, info_name: str):
        """Reads info file (if it exists)
        Args:
            path (Path): Path where info file resides.
            info_name (str): Name of info file.
        """
        # TODO: Allow arbitrary groupings of words
        info_path = path.joinpath(info_name)
        if info_path.exists():
            with open(info_path, 'r', encoding='UTF-8') as info_file:
                # TODO: we don't currently use this. Leaving it be to validate format
                json.load(info_file)

    def _find_audio_format(self, word_dict: Dict[str, Path]) -> str:
        """Determines audio format of voice audio files.

        Args:
            word_dict (Dict[str, Path]): Dict of {filepath: word} associations.

        Raises:
            NoAudioFormatFound: Raised if no audio format can be determined.
            InconsistentAudioFormats: Raised if there are inconsistent audio formats.

        Returns:
            str: Audio format.
        """
        file_format = None
        for path in word_dict.values():
            if file_format is None:
                file_format = path.suffix[1:]
            else:
                if str(file_format) != str(path.suffix[1:]):
                    log.error("Inconsistent audio formats in the word dict. File %s does not match expected format of %s", path, file_format)
                    raise InconsistentAudioFormats
        if not file_format:
            raise NoAudioFormatFound
        log.info("Audio format found: %s", file_format)
        return file_format

    def _get_words(self) -> List[str]:
        """Gets the available words for the voice

        Returns:
            List[str]: Words available to the voice
        """
        word_list = list(self._word_dict.keys())
        word_list.sort()
        return word_list

    def get_audio_format(self) -> str:
        """Get the audio format of the voice files as well as generated files
        Returns:
            (string): Audio format
        """
        return self._audio_format

    def _create_audio_segments(self, word_array: List[str]) -> List[AudioSegment]:
        words_audio: List[AudioSegment] = []
        for word in word_array:
            if word in PUNCTUATION_TIMING:
                words_audio.append(AudioSegment.silent(
                    PUNCTUATION_TIMING[word]))
            else:
                words_audio.append(AudioSegment.from_file(
                    self._word_dict[word], self._audio_format))
        return words_audio


class MultiVoice(Voice):
    """Voice class that uses other voices to assemble
    multi-voice sentences.

    Example: vox:hello hev:there
    Generates a sentence with one word from a voice
    called "vox" and another from a voice called "hev."
    """

    def __init__(self, voices: Dict[str, SingleVoice], database: Optional[DatabaseConnection]):
        """
        Args:
            voices (Dict[str, SingleVoice]): Voices to use to assemble sentences.
            database (Optional[DatabaseConnection]): Database connection information.
                If none provided, no database will be used and no data will persist.
        """
        super().__init__(name='multi', database=database)
        self._voices = voices

        self.words = self._get_words(voices)

    def _get_words(self, voices: Dict[str, SingleVoice]):
        words = []
        for name, voice in voices.items():
            voice_words = [f'{name}:{word}' for word in voice.words]
            words.extend(voice_words)
        return words

    def _get_word_and_voice_strings(self, words_and_voices: List[Tuple[str, SingleVoice]]) -> List[str]:
        """Turns array of word:voice assignments into string.

        Example: "vox:hello vox:there hev:doctor hev:freeman

        Args:
            words_and_voices (List[Tuple[str, SingleVoice]]): Word:voice assignments

        Returns:
            List[str]: Combined string
        """
        return [f'{voice.name}:{word}' for word, voice in words_and_voices]

    def _get_sentence_info(self, words: List[str]) -> Tuple[Sentence, List[str]]:
        # TODO: There is a good amount of double-processing going on here
        words_and_voices = self._get_word_voice_assignment(words)
        words_and_voices_strings = self._get_word_and_voice_strings(
            words_and_voices)
        sayable_words, unsayable_words = self.get_sayable_unsayable(
            words_and_voices_strings)
        sayable_sent_arr = [
            word_voice for word_voice in words_and_voices_strings if word_voice in sayable_words]
        combined_voice_sentences = self.get_combined_voice_sentences(
            words_and_voices)
        sentence_arr = []
        for voice, sentence_words in combined_voice_sentences:
            voice_sentence_segment = f'{voice.name}:{" ".join(sentence_words)}'
            sentence_arr.append(voice_sentence_segment)

        sayable_sent_str = ' '.join(sentence_arr)

        sentence = Sentence(
            sentence=sayable_sent_str,
            sayable=sayable_words,
            unsayable=unsayable_words,
            audio=None,
        )
        return sentence, sayable_sent_arr

    def get_sayable_unsayable(self, words: List[str]) -> Tuple[List[str], List[str]]:
        sayable = []
        unsayable = []
        words_and_voices = self._get_word_voice_assignment(words=words)
        combined_voice_sentences = self.get_combined_voice_sentences(
            words_and_voices)
        for voice, sentence_words in combined_voice_sentences:
            voice_sayable, voice_unsayable = voice.get_sayable_unsayable(sentence_words)
            voice_sayable = [f'{voice.name}:{word}' for word in voice_sayable]
            voice_unsayable = [
                f'{voice.name}:{word}' for word in voice_unsayable]
            sayable.extend(voice_sayable)
            unsayable.extend(voice_unsayable)
        sayable.sort()
        unsayable.sort()
        return sayable, unsayable

    def _create_audio_segments(self, word_array: List[str]) -> List[AudioSegment]:
        words_and_voices = self._get_word_voice_assignment(word_array)
        combined_voice_sentences = self.get_combined_voice_sentences(
            words_and_voices)
        return self.get_combined_audio(
            voice_sentences=combined_voice_sentences,
        )

    def _get_word_voice_assignment(self, words: List[str]) -> List[Tuple[str, SingleVoice]]:
        """Determines voice for each word in a list separated
        from a raw sentence. Only the first word must have a voice
        assignment, further assignments are inferred.

        Example: vox:hello there hev:doctor freeman
        The first two words are assigned to vox, second two to hev

        Args:
            words (List[str]): Words to determine voice assignment of

        Raises:
            FailedToSplit: Raised if unable to split a word/voice assignment.
            NoVoiceSpecified: Raised if initial voice cannot be determined.

        Returns:
            List[Tuple[str, SingleVoice]]: word:voice assignments
        """
        words_and_voices: List[Tuple[str, SingleVoice]] = []

        current_voice: Optional[SingleVoice] = None
        for word_maybe_voice in words:
            word_split = word_maybe_voice.split(':')
            word: Optional[str] = None
            voice: Optional[str] = None
            if len(word_split) == 1:
                word = word_split[0]
            elif len(word_split) == 2:
                voice = word_split[0]
                word = word_split[1]
            if not word:
                raise FailedToSplit

            if voice:
                current_voice = self._voices[voice]
            if not current_voice:
                raise NoVoiceSpecified
            word_and_voice = (word, current_voice)
            words_and_voices.append(word_and_voice)

        return words_and_voices

    def get_combined_voice_sentences(self, words_and_voices: List[Tuple[str, SingleVoice]]) -> List[Tuple[SingleVoice, List[str]]]:
        """Turns individual word:voice assignments into
        combined sentences for each word in sequence:

        Example: vox:hello vox: there hev:doctor hev:freeman vox:boop
        Returns vox:[hello, there] hev:[doctor freeman] vox:[boop]

        Args:
            words_and_voices (List[Tuple[str, SingleVoice]]): Word:voice assignments

        Returns:
            List[Tuple[SingleVoice, List[str]]]: Voice:sentence assignments
        """
        current_voice: Optional[SingleVoice] = None
        current_voice_sentence: List[str] = []
        voice_sentences: List[Tuple[SingleVoice, List[str]]] = []
        for word, voice in words_and_voices:
            if not current_voice:
                current_voice = voice
            if voice == current_voice:
                current_voice_sentence.append(word)
            else:
                voice_sentences.append((current_voice, current_voice_sentence))
                current_voice = voice
                current_voice_sentence = [word]
        if current_voice and current_voice_sentence:
            voice_sentences.append((current_voice, current_voice_sentence))
        return voice_sentences

    def get_combined_audio(self, voice_sentences: List[Tuple[SingleVoice, List[str]]]) -> List[AudioSegment]:
        """Generates audio segments for each voice sentence

        Args:
            voice_sentences (List[Tuple[SingleVoice, List[str]]]): Voice:sentence assignments

        Returns:
            List[AudioSegment]: List of generated audio segments
        """
        audio_segments = []
        for voice, words in voice_sentences:
            sentence = voice.generate_audio_from_array(
                words, save_to_db=False)
            if sentence.audio:
                audio_segments.append(sentence.audio)
        return audio_segments


if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)
    parser = argparse.ArgumentParser(
        description='Generate a sentence using a voice')
    parser.add_argument('-s', '--voice-dir', type=str, required=True,
                        help='Path to folder with voice audio files')
    parser.add_argument('-f', '--format', type=str, required=False,
                        default='wav', help='Audio format to export as')
    parser.add_argument('sentence', type=str)
    args = parser.parse_args()

    voice_dir = Path(args.voice_dir)
    if not voice_dir.is_dir():
        log.error('Voice dir at %s does not exist!', voice_dir)
        sys.exit(1)

    selected_voice = SingleVoice(name=voice_dir.name, path=voice_dir, database=None)
    generated_sentence = selected_voice.generate_audio(args.sentence)
    if generated_sentence is None or generated_sentence.audio is None:
        sys.exit(1)

    output_path = Path.cwd().joinpath(f"{generated_sentence.sentence}.{args.format}")

    generated_sentence.audio.export(output_path, format=args.format)
